{-# LANGUAGE ScopedTypeVariables, FlexibleContexts, TypeFamilies
           , TypeOperators
  #-}
{-# OPTIONS_GHC -Wall -fno-warn-orphans #-}
----------------------------------------------------------------------
-- |
-- Module      :  FRP.Reactive.Behavior
-- Copyright   :  (c) Conal Elliott 2008
-- License     :  BSD3
-- 
-- Maintainer  :  conal@conal.net
-- Stability   :  experimental
-- 
-- Reactive behaviors (continuous time)
----------------------------------------------------------------------

module FRP.Reactive.Behavior
  (
    BehaviorG, Behavior, Behaviour
  , time
  , stepper, switcher --, select
  , snapshotWith, snapshot, snapshot_, whenE
  , accumB, scanlB, monoidB, maybeB, flipFlop, countB
  , sumB, integral
  ) where

import Data.Monoid (Monoid(..))
import Control.Applicative (Applicative,(<$>),pure)
-- import Control.Monad (join)

import Control.Comonad

import Control.Compose ((:.)(..))

import Data.VectorSpace

import qualified FRP.Reactive.Reactive as R
import FRP.Reactive.Reactive (TimeT, ITime, Event, withTimeE, onceRestE, diffE,joinMaybes,result)
import FRP.Reactive.Fun
import FRP.Reactive.Internal.Behavior


-- | Time-specialized behaviors.
-- Note: The signatures of all of the behavior functions can be generalized.  Is
-- the interface generality worth the complexity?
type Behavior = BehaviorG ITime TimeT

-- Synonym for 'Behavior'
type Behaviour = Behavior

-- | The identity generalized behavior.  Has value @t@ at time @t@.
time :: Behavior TimeT
time = beh (pure (fun id))

-- Turn a reactive value into a discretly changing behavior.
rToB :: R.Reactive a -> Behavior a
rToB = beh . fmap pure

-- Then use 'rToB' to promote reactive value functions to behavior
-- functions.

-- | Discretely changing behavior, based on an initial value and a
-- new-value event.
stepper :: a -> Event a -> Behavior a
stepper = (result.result) rToB R.stepper

-- Suggested by Robin Green:

-- stepper = select pure

-- -- | Use a key event to key into a behaviour-valued function
-- select :: (a -> Behavior b) -> a -> Event a -> Behavior b
-- select f a e = f a `switcher` (f <$> e)

-- Looking for a more descriptive name.

-- | Switch between behaviors.
switcher :: Behavior a -> Event (Behavior a) -> Behavior a
b `switcher` eb = beh (unb b `R.switcher` (unb <$> eb))

-- | Snapshots a behavior whenever an event occurs and combines the values
-- using the combining function passed.
snapshotWith :: (a -> b -> c) -> Event a -> Behavior b -> Event c
snapshotWith h e b = f <$> (withTimeE e `R.snapshot` unb b)
 where
   f ((a,t),tfun) = h a (tfun `apply` t)

-- | Snapshot a behavior whenever an event occurs.  See also 'snapshotWith'.
snapshot :: Event a -> Behavior b -> Event (a,b)
snapshot = snapshotWith (,)

-- Alternative implementations:
--   snapshotWith c e b = uncurry c <$> snapshot e b
--   snapshotWith c = (result.result.fmap) (uncurry c) snapshot

-- | Like 'snapshot' but discarding event data (often @a@ is '()').
snapshot_ :: Event a -> Behavior b -> Event b
snapshot_ = snapshotWith (flip const)

-- Alternative implementations
-- e `snapshot_` src = snd <$> (e `snapshot` src)
-- snapshot_ = (result.result.fmap) snd snapshot

-- | Filter an event according to whether a reactive boolean is true.
whenE :: Event a -> Behavior Bool -> Event a
whenE e = joinMaybes . fmap h . snapshot e
 where
   h (a,True)  = Just a
   h (_,False) = Nothing

-- | Behavior from an initial value and an updater event.  See also
-- 'accumE'.
accumB :: a -> Event (a -> a) -> Behavior a
accumB = (result.result) rToB R.accumR

-- -- | Like 'scanl' for behaviors.  See also 'scanlE'.
-- scanlB :: (a -> b -> a) -> a -> Event b -> Behavior a
-- scanlB = (result.result.result) rToB R.scanlR

-- -- | Accumulate values from a monoid-valued event.  Specialization of
-- -- 'scanlB', using 'mappend' and 'mempty'.  See also 'monoidE'.
-- monoidB :: Monoid a => Event a -> Behavior a
-- monoidB = result rToB R.monoidR


---- The next versions are more continuous:

-- type RF a = R.Reactive (Fun TimeT a)

-- scanlB :: forall a c. (Behavior a -> c -> Behavior a) -> Behavior a
--        -> Event c -> Behavior a
-- scanlB f b0 e = beh (scanlRF f' (unb b0) e)
--  where
--    f' :: RF a -> c -> RF a
--    f' r c = unb (f (beh r) c)

-- scanlRF :: (RF a -> c -> RF a) -> RF a -> Event c -> RF a
-- scanlRF h rf0 e = join (R.scanlR h rf0 e)

-- monoidB :: Monoid a => Event (Behavior a) -> Behavior a
-- monoidB = scanlB mappend mempty

-- -- I doubt these definitions work well.  They accumulate reactives without
-- -- aging them.  See 'accumE'.


-- | Like 'scanl' for behaviors.  See also 'scanlE'.
scanlB :: forall a. (Behavior a -> Behavior a -> Behavior a) -> Behavior a
       -> Event (Behavior a) -> Behavior a
scanlB plus zero = h
 where
   h :: Event (Behavior a) -> Behavior a
   h e = zero `switcher` (g <$> onceRestE e)
   g :: (Behavior a, Event (Behavior a)) -> Behavior a
   g (b, e') = b `plus` h e'

-- | Accumulate values from a monoid-valued event.  Specialization of
-- 'scanlB', using 'mappend' and 'mempty'.  See also 'monoidE'.
monoidB :: Monoid a => Event (Behavior a) -> Behavior a
monoidB = scanlB mappend mempty

-- | Like 'sum' for behaviors.
sumB :: AdditiveGroup a => Event a -> Behavior a
sumB = result rToB R.sumR

-- | Start out blank ('Nothing'), latching onto each new @a@, and blanking
-- on each @b@.  If you just want to latch and not blank, then use
-- 'mempty' for the second event.
maybeB :: Event a -> Event b -> Behavior (Maybe a)
maybeB = (result.result) rToB R.maybeR

-- | Flip-flopping behavior.  Turns true whenever first event occurs and
-- false whenever the second event occurs.
flipFlop :: Event a -> Event b -> Behavior Bool
flipFlop = (result.result) rToB R.flipFlop

-- | Count occurrences of an event.  See also 'countE'.
countB :: Num n => Event a -> Behavior n
countB = result rToB R.countR

-- | Euler integral.
integral :: (VectorSpace v, Scalar v ~ TimeT) =>
            Event () -> Behavior v -> Behavior v
integral t = sumB . snapshotWith (*^) (diffE (t `snapshot_` time))

-- TODO: find out whether this integral works recursively.  If not, then
-- fix the implementation, rather than changing the semantics.  (No
-- "delayed integral".)
-- 
-- Early experiments suggest that recursive integration gets stuck.
-- Investigate.


---- Comonadic stuff

-- Orphan.  Move elsewhere

instance (Functor g, Functor f, Copointed g, Copointed f)
      => Copointed (g :. f) where
  extract = extract . fmap extract . unO

-- WORKING HERE

-- The plan for duplicate:
--
--   (g :. f) a -> g (f a) -> g (f (f a)) -> g (g (f (f a)))
--              -> g (f (g (f a))) -> (g :. f) (g (f a))
--              -> (g :. f) ((g :. f) a) -> 

-- But we'll have to do that middle twiddle, which I couldn't do for
-- behaviors to get a Monad either.  Is there another way?


-- instance Comonad (g :. f) where
--   duplicate 



-- Standard instances for applicative functors

-- #define APPLICATIVE Behavior
-- #include "Num-inc.hs"
